#include "SDL.h"
#include "SDL_image.h"
#include <stdio.h>
#include <string.h>

#include "ehash.h"
#include "ratio.h"
#include "bsicon.h"
#include "SDL_savejpeg.h"
#include "SDL_subsurface.h"
//#include "SDL_rotozoom.h"

void draw_box(SDL_Surface *s, fraction *ratio, SDL_Rect *area2);
int scale_max(const SDL_Surface *s, int maxw, int maxh, fraction *ratio);
void rect_to_points(SDL_Rect *area);
void points_to_rect(SDL_Rect *points);

#define SSTACK_MAX 32

struct
{
  int x,y;
  int quality;
  const char *postfix;
} scalings[]={{800,600,85,NULL         },
              {800,600,85,"n.jpg"      },
              {150, -1,-1,"n-thumb.jpg"},
              { -1, -1,-1,NULL         } };
enum scale_e
{
  SCALE_DISPLAY,
  SCALE_RESIZED,
  SCALE_THUMBNAIL
};

#define SCALE_MAXW (scalings[SCALE_RESIZED].x)
#define SCALE_MAXH (scalings[SCALE_RESIZED].y)
#define DISPL_MAXW (scalings[SCALE_DISPLAY].x)
#define DISPL_MAXH (scalings[SCALE_DISPLAY].y)
#define THUMB_MAXW (scalings[SCALE_THUMBNAIL].x)
#define THUMB_MAXH (scalings[SCALE_THUMBNAIL].y)

SDL_Surface *SDL_SoftResize(SDL_Surface *s, int w, int h)
{
  SDL_Rect srcrect={0,0,s->w,s->h},
           dstrect={0,0,w,h};
  SDL_Surface *sout=SDL_CreateRGBSurface(SDL_SWSURFACE,w,h,
    s->format->BitsPerPixel,
    s->format->Rmask,
    s->format->Gmask,
    s->format->Bmask,
    s->format->Amask);

  if(sout==NULL)
    goto SDL_SOFTRESIZE_ERR;
  if(SDL_SoftStretch(s,&srcrect,sout,&dstrect)<0)
    goto SDL_SOFTRESIZE_ERR;

  return(sout);
SDL_SOFTRESIZE_ERR:
  if(sout != NULL) SDL_FreeSurface(sout);
  return(NULL);
}

int save_with_thumbnail(SDL_Surface *s, const char *source, const char *opath)
{
  int n;
  if(opath == NULL) opath="./";

  for(n=SCALE_RESIZED; n<=SCALE_THUMBNAIL; n++)
  {
    char buf[1024],*ext;
    int pos;
    fraction ratio={1,1};
    double scale;

    ext=strrchr(source,'/');
    if(ext != NULL) pos=sprintf(buf,"%s/%s",opath,ext+1);
    else            pos=sprintf(buf,"%s/%s",opath,source);    

    fprintf(stderr,"name w/o replaced extensions:  %s\n",buf);

    ext=strrchr(buf,'.');
    if(ext != NULL) strcpy(ext,scalings[n].postfix);
    else            strcat(buf+pos,scalings[n].postfix);

    fprintf(stderr,"name w/replaced extensions:  %s\n",buf);

    scale_max(s,scalings[n].x,scalings[n].y,&ratio);
    scale=((double)ratio.numerator)/((double)(ratio.denominator));
    fprintf(stderr,"Scaling factor %.2f\n",scale);
    if(ratio.numerator < ratio.denominator)
    {
      SDL_Surface *s2;
      SDL_Surface srcrect={0,0,s->w,s->h};
      int w=RATIO_MUL(s->w,&ratio);
      int h=RATIO_MUL(s->h,&ratio);

      fprintf(stderr,"Scaling (%d,%d)->(%d,%d)\n",
        srcrect.w,srcrect.h,w,h);

      s2=SDL_SoftResize(s,w,h);
//      s2=rotozoomSurface(s,0.0f,scale,1);
//      fprintf(stderr,"Scaled to (%d,%d)\n",s2->w,s2->h);

      SDL_SaveJPEG(s2,buf,scalings[n].quality);
      SDL_FreeSurface(s2);
    }
    else
      SDL_SaveJPEG(s,buf,scalings[n].quality);
  }
}

int main(int argc, char *argv[])
{
  const char *odir=NULL;
  SDL_Surface *screen=NULL;
  union
  {
    SDL_Surface *base;
    SDL_Surface *array[SSTACK_MAX];
  } sstack;
  int sstackpos=-1,running=1;
  int n;
  int startpos=0;
  int proc_args_add;

  if(argc < 2)
  {
    fprintf(stderr,"syntax:  photocrop -o dir file1 [file2] [...]\n");
    return(2);
  }

  if(SDL_Init(SDL_INIT_VIDEO)<0)
  {
    fprintf(stderr,"Couldn't init SDL: %s\n",SDL_GetError());
    return(1);
  }

  proc_args_add=1;
  while(proc_args_add > 0)
  {
    startpos+=proc_args_add;
    proc_args_add=0;

    switch(argv[startpos][0])
    {
    case '-':
      switch(argv[startpos][1])
      {
      case 'o':
        if(argv[startpos][2] != '\0')
        {
          fprintf(stderr,"Unknown argument %s\n",argv[startpos]);
          return(1);
        }
        odir=argv[startpos+1];
        proc_args_add=2;
        break;

      /* -X:  window width -Y:  window height */
      case 'X':
        if(argv[startpos][2] != '\0')
        {
          fprintf(stderr,"Unknown argument %s\n",argv[startpos]);
          return(1);
        }

        DISPL_MAXW=atoi(argv[startpos+1]);
        if(DISPL_MAXW <= 0)
        {
          fprintf(stderr,"Invalid display max width %s\n",argv[startpos+1]);
          return(1);
        }
        proc_args_add=2;
        break;

      case 'Y':        
        if(argv[startpos][2] != '\0')
        {
          fprintf(stderr,"Unknown argument %s\n",argv[startpos]);
          return(1);
        }

        DISPL_MAXH=atoi(argv[startpos+1]);
        if(DISPL_MAXH <= 0)
        {
          fprintf(stderr,"Invalid display max height %s\n",argv[startpos+1]);
          return(1);
        }
        proc_args_add=2;
        break;

      case 'W':
        if(argv[startpos][2] != '\0')
        {
          fprintf(stderr,"Unknown argument %s\n",argv[startpos]);
          return(1);
        }

        SCALE_MAXW=atoi(argv[startpos+1]);
        if(SCALE_MAXW <= 0)
        {
          fprintf(stderr,"Invalid resized max width %s\n",argv[startpos+1]);
          return(1);
        }
        proc_args_add=2;
        break;

      case 'H':
        if(argv[startpos][2] != '\0')
        {
          fprintf(stderr,"Unknown argument %s\n",argv[startpos]);
          return(1);
        }

        SCALE_MAXH=atoi(argv[startpos+1]);
        if(SCALE_MAXH <= 0)
        {
          fprintf(stderr,"Invalid resized max height %s\n",argv[startpos+1]);
          return(1);
        }
        proc_args_add=2;
        break;

      case 'w':
        if(argv[startpos][2] != '\0')
        {
          fprintf(stderr,"Unknown argument %s\n",argv[startpos]);
          return(1);
        }

        THUMB_MAXW=atoi(argv[startpos+1]);
        if(THUMB_MAXW <= 0)
        {
          fprintf(stderr,"Invalid thumbnail max width %s\n",argv[startpos+1]);
          return(1);
        }
        proc_args_add=2;
        break;

      case 'h':
        if(argv[startpos][2] != '\0')
        {
          fprintf(stderr,"Unknown argument %s\n",argv[startpos]);
          return(1);
        }

        THUMB_MAXH=atoi(argv[startpos+1]);
        if(THUMB_MAXH <= 0)
        {
          fprintf(stderr,"Invalid thumbnail max height %s\n",argv[startpos+1]);
          return(1);
        }
        proc_args_add=2;
        break;

      default:
        break;
      }
    default:
      break;
    }

/*      if(strcmp(argv[startpos],"-o") == 0)
      {
        odir=argv[startpos+1];
        proc_args_add=2;
        proc_args_add=1;
      }*/
  }

  for(n=startpos; (running>=0)&&(n<argc); n++)
  {
    SDL_Event event;
    int redraw=1,resize=1;
    fraction ratio={1,1};
    SDL_Rect scwidth={0,0},dstwidth={0,0},croparea={0,0,0,0};

    event.type=0;

    while(sstackpos >= 0)
      SDL_FreeSurface(sstack.array[sstackpos--]);

    if(screen == NULL)
    {
      BS_SetIcon();
      screen=SDL_SetVideoMode(320,240,32,0);
    }

    sstack.base=IMG_Load(argv[n]);
    if(sstack.base == NULL)
    {
      fprintf(stderr,"Couldn't load %s: %s\n",argv[n],SDL_GetError());
      continue;
    }
    {
      SDL_Surface *img2=SDL_DisplayFormat(sstack.base);
      SDL_FreeSurface(sstack.base);
      sstack.base=img2;
    }
    sstackpos++;

    running=1;

    do 
    {
      if(resize)
      {
        scwidth.w=sstack.array[sstackpos]->w;
        scwidth.h=sstack.array[sstackpos]->h;
        dstwidth=scwidth;
        croparea=scwidth;

        scale_max(sstack.array[sstackpos],DISPL_MAXW,DISPL_MAXH,&ratio);
        ratio_mul(&ratio,&dstwidth);

        screen=SDL_SetVideoMode(dstwidth.w,dstwidth.h,32,0);
        if(screen==NULL)
        {
          fprintf(stderr,"Couldn't set video mode: %s\n",SDL_GetError());
          continue;
        }

        {
          static const char *prefix="Photocrop";
          char buf[256];
          sprintf(buf,"%s: [%s]",prefix,argv[n]);
          SDL_WM_SetCaption(buf,prefix);
        }
        resize=0;
      }

      if(redraw)
      {
        if(SDL_SoftStretch(sstack.array[sstackpos],&scwidth,screen,&dstwidth)<0)
          fprintf(stderr,"failed: %s\n",SDL_GetError());
        draw_box(screen,&ratio,&croparea);
        SDL_Flip(screen);
        redraw=0;
      }

      switch(EventHash(&event))
      {
      int scalex,scaley;
      case SDL_QUIT:
        running=-1;
        break;

      case SDL_KEYDOWN_H(SDLK_RETURN,KMOD_CTRL):
      case SDL_KEYDOWN_H(SDLK_KP_ENTER,KMOD_CTRL):
        {
          SDL_Surface *s=SDL_SubSurface(sstack.array[sstackpos],
            &croparea);
          fprintf(stderr,"Created subsurface...\n");
          save_with_thumbnail(s,argv[n],odir);
          fprintf(stderr,"Saved subsurface...\n");
          SDL_FreeSurface(s);
          fprintf(stderr,"Freed subsurface...\n");
          running=0;
        }
        continue;

      case SDL_KEYDOWN_H(SDLK_HOME,0):
      case SDL_KEYDOWN_H(SDLK_KP7,0):
        if(n!=1)
        {
          n=0;
          running=0;
        }
        continue;

      case SDL_KEYDOWN_H(SDLK_END,0):
      case SDL_KEYDOWN_H(SDLK_KP1,0):
        if((argc>=2) && (n != (argc-1)))
        {
          n=(argc-2);
          running=0;
        }
        break;

      case SDL_KEYDOWN_H(SDLK_PAGEUP,0):
      case SDL_KEYDOWN_H(SDLK_KP9,0):
        if(n > 2)
        {
          n-=2;
          running=0;
        }
        continue;

      case SDL_KEYDOWN_H(SDLK_PAGEDOWN,0):
      case SDL_KEYDOWN_H(SDLK_KP3,0):
        if(n < (argc-1))
          running=0;
        continue;

      case SDL_KEYDOWN_H(SDLK_KP_MULTIPLY,0):
        croparea=scwidth;
        redraw=1;
        break;

      case SDL_KEYDOWN_H(SDLK_KP_MINUS,0):
        if(sstackpos > 0)
        {
          SDL_FreeSurface(sstack.array[sstackpos--]);
          SDL_UnlockSurface(sstack.array[sstackpos]);
          redraw=1;
          resize=1;
        }
        break;

      case SDL_KEYDOWN_H(SDLK_KP_PLUS,0):
        if((sstackpos+1)<SSTACK_MAX)
        {
          sstack.array[sstackpos+1]=
            SDL_SubSurface(sstack.array[sstackpos],&croparea);
          sstackpos++;
          croparea.x=0;
          croparea.y=0;
          scwidth.w=croparea.w;
          scwidth.h=croparea.h;
          redraw=1;
          resize=1;
        }
        break;

      case SDL_KEYDOWN_H(SDLK_UP,0):
        rect_to_points(&croparea);
          if(croparea.y > 0)
            croparea.y--;
        points_to_rect(&croparea);
        redraw=1;
        break;

      case SDL_KEYDOWN_H(SDLK_DOWN,0):
        rect_to_points(&croparea);
          croparea.y++;
        points_to_rect(&croparea);
        redraw=1;
        break;

      case SDL_KEYDOWN_H(SDLK_LEFT,0):
        rect_to_points(&croparea);
          if(croparea.x > 0)
            croparea.x--;
        points_to_rect(&croparea);
        redraw=1;
        break;

      case SDL_KEYDOWN_H(SDLK_RIGHT,0):
        rect_to_points(&croparea);
          croparea.x++;
        points_to_rect(&croparea);
        redraw=1;
        break;

      case SDL_KEYDOWN_H(SDLK_UP,KMOD_CTRL):
        rect_to_points(&croparea);
          if(croparea.h > 0)
            croparea.h--;
        points_to_rect(&croparea);
        redraw=1;
        break;

      case SDL_KEYDOWN_H(SDLK_DOWN,KMOD_CTRL):
        rect_to_points(&croparea);
          croparea.h++;
        points_to_rect(&croparea);
        redraw=1;
        break;

      case SDL_KEYDOWN_H(SDLK_LEFT,KMOD_CTRL):
        rect_to_points(&croparea);
          if(croparea.w > 0)
            croparea.w--;
        points_to_rect(&croparea);
        redraw=1;
        break;

      case SDL_KEYDOWN_H(SDLK_RIGHT,KMOD_CTRL):
        rect_to_points(&croparea);
          croparea.w++;
        points_to_rect(&croparea);
        redraw=1;
        break;

      case SDL_MOUSEDOWN_H(SDL_BUTTON_LEFT,0):
        scalex=RATIO_DIV(event.button.x,&ratio);
        scaley=RATIO_DIV(event.button.y,&ratio);

        rect_to_points(&croparea);
          if((scalex < croparea.w) && (scaley < croparea.h))
          {
            croparea.x=RATIO_DIV(event.button.x,&ratio);
            croparea.y=RATIO_DIV(event.button.y,&ratio);
          }
        points_to_rect(&croparea);
        redraw=1;
        break;

      case SDL_MOUSEDOWN_H(SDL_BUTTON_RIGHT,0):
        scalex=RATIO_DIV(event.button.x,&ratio);
        scaley=RATIO_DIV(event.button.y,&ratio);

        if((scalex > croparea.x) && (scaley > croparea.y))
        {
          rect_to_points(&croparea);
            croparea.w=RATIO_DIV(event.button.x,&ratio);
            croparea.h=RATIO_DIV(event.button.y,&ratio);
          points_to_rect(&croparea);
          redraw=1;
        }
        break;

      default:
        break;
      }
    } while((running>0) && SDL_WaitEvent(&event));
  }

  while(sstackpos >= 0)
    SDL_FreeSurface(sstack.array[sstackpos--]);

  SDL_Quit();
  return(0);
}

int scale_max(const SDL_Surface *s, int maxw, int maxh,
  fraction *ratio)
{
  fraction xratio={-1,0},yratio={-1,0};

  if((maxw<=0)&&(maxh<=0))
  {
    SDL_SetError("Invalid (maxw,maxh)=(%d,%d)",maxw,maxh);
    return(-1);
  }

  xratio.numerator=maxw;
  xratio.denominator=s->w;
  yratio.numerator=maxh;
  yratio.denominator=s->h;

  if(yratio.numerator < 0)
    (*ratio)=xratio;
  else if(xratio.numerator < 0)
    (*ratio)=yratio;
  else if(ratio_greater(&yratio,&xratio))
    (*ratio)=xratio;
  else
    (*ratio)=yratio;

  return(0);
}

void rect_to_points(SDL_Rect *area)
{
  area->w += area->x;
  area->h += area->y;
}

void points_to_rect(SDL_Rect *points)
{
  if(points->x > points->w)
  {
    int tmp=points->x;
    points->x=points->w;
    points->w=tmp;
  }

  if(points->y > points->h)
  {
    int tmp=points->y;
    points->y=points->h;
    points->h=tmp;
  }

  points->w -= points->x;
  points->h -= points->y;
}

#define CORNER_W 8
#define CORNER_H 8

void draw_box(SDL_Surface *s, fraction *ratio, SDL_Rect *area2)
{
  Uint32 color=SDL_MapRGB(s->format,255,255,0);
  SDL_Rect area=(*area2);

  ratio_mul(ratio,&area);
  {
    SDL_Rect tlcorner=  {area.x,
                         area.y,
                         CORNER_W,
                         CORNER_H};
    SDL_Rect lrcorner=  {(area.x + area.w)-CORNER_W,
                         (area.y + area.h)-CORNER_H,
                         CORNER_W,
                         CORNER_H};
    SDL_Rect topline=   {area.x+CORNER_W,
                         area.y,
                         area.w-CORNER_W,
                         1       };
    SDL_Rect bottomline={area.x,
                         area.y+(area.h-1),
                         area.w-CORNER_W,
                         1       };
    SDL_Rect leftline=  {area.x,
                         area.y+CORNER_H,
                         1,
                         area.h-CORNER_H};
    SDL_Rect rightline=  {area.x+(area.w-1),
                         area.y,
                         1,
                         area.h-CORNER_H};

    SDL_FillRect(s,&topline,color);
    SDL_FillRect(s,&bottomline,color);
    SDL_FillRect(s,&leftline,color);
    SDL_FillRect(s,&rightline,color);
    SDL_FillRect(s,&tlcorner,color);
    SDL_FillRect(s,&lrcorner,color);
  }
}
